@@ -18,6 +18,12 @@
 #
 ####################################################################################################
 
+# Fixme: versus PySpice.Plot ???
+
+####################################################################################################
+
+"""This module implements plotting helper."""
+
 ####################################################################################################
 
 import matplotlib.pyplot as plt
@@ -35,9 +41,3 @@ def plot(waveform, *args, **kwargs):
     if 'axis' in kwargs:
         del kwargs['axis']
     axis.plot(waveform.abscissa, waveform, *args, **kwargs)
-
-####################################################################################################
-#
-# End
-#
-####################################################################################################
