@@ -20,22 +20,37 @@
 #
 ####################################################################################################
 
-""" This module implements high level elements built on top of Spice elements. """
+"""This module implements high level elements built on top of Spice elements."""
 
-# Fixme: these waveforms can be current sources as well
+# Fixme: check NgSpice for discrepancies
 
 ####################################################################################################
 
 from ..Math import rms_to_amplitude, amplitude_to_rms
-from ..Tools.StringTools import join_list, join_dict, str_spice
+from ..Tools.StringTools import join_list, join_dict, str_spice, str_spice_list
 from ..Unit import as_s, as_V, as_A, as_Hz
-from .BasicElement import VoltageSource
+from .BasicElement import VoltageSource, CurrentSource
 
 ####################################################################################################
 
-class Sinusoidal(VoltageSource):
+class SourceMixinAbc:
+    AS_UNIT = None
 
-    r""" This class implements a sinusoidal waveform.
+####################################################################################################
+
+class VoltageSourceMixinAbc:
+    AS_UNIT = as_V
+
+####################################################################################################
+
+class CurrentSourceMixinAbc:
+    AS_UNIT = as_A
+
+####################################################################################################
+
+class SinusoidalMixin(SourceMixinAbc):
+
+    r"""This class implements a sinusoidal waveform.
 
     +------+----------------+---------------+-------+
     | Name + Parameter      + Default Value + Units |
@@ -66,6 +81,8 @@ class Sinusoidal(VoltageSource):
 
     Public Attributes:
 
+      :attr:`ac_magnitude`
+
       :attr:`amplitude`
 
       :attr:`damping_factor`
@@ -82,17 +99,17 @@ class Sinusoidal(VoltageSource):
 
     ##############################################
 
-    def __init__(self, name, node_plus, node_minus,
+    def __init__(self,
                  dc_offset=0,
+                 ac_magnitude=1,
                  offset=0, amplitude=1, frequency=50,
                  delay=0, damping_factor=0):
 
-        super().__init__(name, node_plus, node_minus)
-
-        self.dc_offset = as_V(dc_offset)
-        self.offset = as_V(offset)
-        self.amplitude = as_V(amplitude)
-        self.frequency = as_Hz(frequency) # Fixme: protect by setter?
+        self.dc_offset = self.AS_UNIT(dc_offset)
+        self.ac_magnitude = self.AS_UNIT(ac_magnitude)
+        self.offset = self.AS_UNIT(offset)
+        self.amplitude = self.AS_UNIT(amplitude)
+        self.frequency = as_Hz(frequency)   # Fixme: protect by setter?
         self.delay = as_s(delay)
         self.damping_factor = as_Hz(damping_factor)
 
@@ -100,7 +117,8 @@ class Sinusoidal(VoltageSource):
 
     @property
     def rms_voltage(self):
-        return amplitude_to_rms(self.amplitude)
+        # Fixme: ok ???
+        return amplitude_to_rms(self.amplitude * self.ac_magnitude)
 
     ##############################################
 
@@ -111,28 +129,15 @@ class Sinusoidal(VoltageSource):
     ##############################################
 
     def format_spice_parameters(self):
-
         sin_part = join_list((self.offset, self.amplitude, self.frequency, self.delay, self.damping_factor))
         return join_list((
-            'DC {}'.format(str_spice(self.dc_offset)),
-            'AC SIN({})'.format(sin_part),
+            'DC {} AC {}'.format(*str_spice_list(self.dc_offset, self.ac_magnitude)),
+            'SIN({})'.format(sin_part),
         ))
 
 ####################################################################################################
 
-class AcLine(Sinusoidal):
-
-    ##############################################
-
-    def __init__(self, name, node_plus, node_minus, rms_voltage=230, frequency=50):
-
-        super().__init__(name, node_plus, node_minus,
-                         amplitude=rms_to_amplitude(rms_voltage),
-                         frequency=frequency)
-
-####################################################################################################
-
-class Pulse(VoltageSource):
+class PulseMixin(SourceMixinAbc):
 
     """This class implements a pulse waveform.
 
@@ -155,10 +160,14 @@ class Pulse(VoltageSource):
     +--------+---------------+---------------+-------+
     | Period + period        + Tstop         + sec   |
     +--------+---------------+---------------+-------+
+    | Phase  + phase         + 0.0           + sec   |
+    +--------+---------------+---------------+-------+
+
+    Phase is only possible when XSPICE is enabled
 
     Spice Syntax::
 
-        PULSE ( V1 V2 Td Tr Tf Pw Period )
+        PULSE ( V1 V2 Td Tr Tf Pw Period Phase )
 
     A single pulse so specified is described by the following table:
 
@@ -191,6 +200,8 @@ class Pulse(VoltageSource):
 
       :attr:`period`
 
+      :attr:`phase`
+
       :attr:`pulse_width`
 
       :attr:`pulsed_value`
@@ -201,24 +212,31 @@ class Pulse(VoltageSource):
 
     ##############################################
 
-    def __init__(self, name, node_plus, node_minus,
+    def __init__(self,
                  initial_value, pulsed_value,
                  pulse_width, period,
-                 delay_time=0, rise_time=0, fall_time=0):
+                 delay_time=0, rise_time=0, fall_time=0,
+                 phase=None,
+                 dc_offset=0):
 
         # Fixme: default
         #  rise_time, fall_time = Tstep
         #  pulse_width, period = Tstop
 
-        super().__init__(name, node_plus, node_minus)
-
-        self.initial_value = as_V(initial_value)
-        self.pulsed_value = as_V(pulsed_value)
+        self.dc_offset = self.AS_UNIT(dc_offset)   # Fixme: -> SourceMixinAbc
+        self.initial_value = self.AS_UNIT(initial_value)
+        self.pulsed_value = self.AS_UNIT(pulsed_value)
         self.delay_time = as_s(delay_time)
         self.rise_time = as_s(rise_time)
         self.fall_time = as_s(fall_time)
         self.pulse_width = as_s(pulse_width)
-        self.period = as_s(period) # Fixme: protect by setter?
+        self.period = as_s(period)   # Fixme: protect by setter?
+
+        # XSPICE
+        if phase is not None:
+            self.phase = as_s(phase)
+        else:
+            self.phase = None
 
         # # Fixme: to func?
         # # Check parameters
@@ -241,15 +259,21 @@ class Pulse(VoltageSource):
 
     def format_spice_parameters(self):
 
+        # if DC is not provided, ngspice complains
+        #   Warning: vpulse: no DC value, transient time 0 value used
+
         # Fixme: to func?
-        return ('PULSE(' +
-                join_list((self.initial_value, self.pulsed_value, self.delay_time,
-                           self.rise_time, self.fall_time, self.pulse_width, self.period)) +
-                ')')
+        return join_list((
+            'DC {}'.format(str_spice(self.dc_offset)),
+            'PULSE(' +
+            join_list((self.initial_value, self.pulsed_value, self.delay_time,
+                       self.rise_time, self.fall_time, self.pulse_width, self.period,
+                       self.phase)) +
+            ')'))
 
 ####################################################################################################
 
-class Exponential(VoltageSource):
+class ExponentialMixin(SourceMixinAbc):
 
     r"""This class implements a Exponential waveform.
 
@@ -293,18 +317,15 @@ class Exponential(VoltageSource):
 
     ##############################################
 
-    def __init__(self, name, node_plus, node_minus,
+    def __init__(self,
                  initial_value, pulsed_value,
                  rise_delay_time=.0, rise_time_constant=None,
-                 fall_delay_time=None, fall_time_constant=None,
-             ):
+                 fall_delay_time=None, fall_time_constant=None):
 
         # Fixme: default
 
-        super().__init__(name, node_plus, node_minus)
-
-        self.initial_value = as_V(initial_value)
-        self.pulsed_value = as_V(pulsed_value)
+        self.initial_value = self.AS_UNIT(initial_value)
+        self.pulsed_value = self.AS_UNIT(pulsed_value)
         self.rise_delay_time = as_s(rise_delay_time)
         self.rise_time_constant = as_s(rise_time_constant)
         self.fall_delay_time = as_s(fall_delay_time)
@@ -313,18 +334,17 @@ class Exponential(VoltageSource):
     ##############################################
 
     def format_spice_parameters(self):
-
         # Fixme: to func?
         return ('EXP(' +
                 join_list((self.initial_value, self.pulsed_value,
                            self.rise_delay_time, self.rise_time_constant,
                            self.fall_delay_time, self.fall_time_constant,
-                       )) +
+                           )) +
                 ')')
 
 ####################################################################################################
 
-class PieceWiseLinear(VoltageSource):
+class PieceWiseLinearMixin(SourceMixinAbc):
 
     r"""This class implements a Piece-Wise Linear waveform.
 
@@ -342,36 +362,49 @@ class PieceWiseLinear(VoltageSource):
     delay time time = td. The current source still needs to be patched, td and r are not yet
     available.
 
+    `values` should be given as a list of (`Time`, `Value`)-tuples, e.g.::
+
+        PieceWiseLinearVoltageSource(
+            circuit,
+            'pwl1', '1', '0',
+            values=[(0, 0), (10@u_ms, 0), (11@u_ms, 5@u_V), (20@u_ms, 5@u_V)],
+        )
+
     """
 
     ##############################################
 
-    def __init__(self, name, node_plus, node_minus,
-                 values,
-                 repeate_time=0, delay_time=.0,
-             ):
-
-        # Fixme: default
-
-        super().__init__(name, node_plus, node_minus)
-
-        self.values = [as_V(x) for x in values]
-        self.repeate_time = as_s(repeate_time)
-        self.delay_time = as_s(delay_time)
+    def __init__(self, values, repeat_time=None, delay_time=None, dc=None):
+        self.values = sum(([as_s(t), self.AS_UNIT(x)] for (t, x) in values), [])
+        self.repeat_time = as_s(repeat_time, none=True)
+        self.delay_time = as_s(delay_time, none=True)
+        self.dc = self.AS_UNIT(dc, none=True)
 
     ##############################################
 
     def format_spice_parameters(self):
 
         # Fixme: to func?
-        return ('PWL(' +
-                join_list(self.values) +
-                join_dict({'r':self.repeate_time, 'td':self.delay_time}) + # OrderedDict(
-                ')')
+
+        d = {}
+        if self.repeat_time is not None:
+            d["r"] = self.repeat_time
+        if self.delay_time is not None:
+            d["td"] = self.delay_time
+
+        _ = ""
+        if self.dc is not None:
+            _ += "DC {} ".format(str_spice(self.dc))
+        _ += "PWL(" + join_list(self.values)
+        if d:
+            _ += " " + join_dict(d)   # OrderedDict(
+        _ += ")"
+
+        return _
 
 ####################################################################################################
 
-class SingleFrequencyFM(VoltageSource):
+class SingleFrequencyFMMixin(SourceMixinAbc):
 
     r"""This class implements a Single-Frequency FM waveform.
 
@@ -403,13 +436,9 @@ class SingleFrequencyFM(VoltageSource):
 
     ##############################################
 
-    def __init__(self, name, node_plus, node_minus,
-                 offset, amplitude, carrier_frequency, modulation_index, signal_frequency):
-
-        super().__init__(name, node_plus, node_minus)
-
-        self.offset = as_V(offset)
-        self.amplitude = as_V(amplitude)
+    def __init__(self, offset, amplitude, carrier_frequency, modulation_index, signal_frequency):
+        self.offset = self.AS_UNIT(offset)
+        self.amplitude = self.AS_UNIT(amplitude)
         self.carrier_frequency = as_Hz(carrier_frequency)
         self.modulation_index = modulation_index
         self.signal_frequency = as_Hz(signal_frequency)
@@ -417,7 +446,6 @@ class SingleFrequencyFM(VoltageSource):
     ##############################################
 
     def format_spice_parameters(self):
-
         # Fixme: to func?
         return ('SFFM(' +
                 join_list((self.offset, self.amplitude, self.carrier_frequency,
@@ -426,7 +454,7 @@ class SingleFrequencyFM(VoltageSource):
 
 ####################################################################################################
 
-class AmplitudeModulated(VoltageSource):
+class AmplitudeModulatedMixin(SourceMixinAbc):
 
     r"""This class implements a Amplitude Modulated source.
 
@@ -458,15 +486,12 @@ class AmplitudeModulated(VoltageSource):
 
     ##############################################
 
-    def __init__(self, name, node_plus, node_minus,
-                 offset, amplitude, modulating_frequency, carrier_frequency, signal_delay):
+    def __init__(self, offset, amplitude, modulating_frequency, carrier_frequency, signal_delay):
 
         # Fixme: default
 
-        super().__init__(name, node_plus, node_minus)
-
-        self.offset = as_V(offset)
-        self.amplitude = as_V(amplitude)
+        self.offset = self.AS_UNIT(offset)
+        self.amplitude = self.AS_UNIT(amplitude)
         self.carrier_frequency = as_Hz(carrier_frequency)
         self.modulating_frequency = as_Hz(modulating_frequency)
         self.signal_delay = as_s(signal_delay)
@@ -474,7 +499,6 @@ class AmplitudeModulated(VoltageSource):
     ##############################################
 
     def format_spice_parameters(self):
-
         # Fixme: to func?
         return ('AM(' +
                 join_list((self.offset, self.amplitude, self.carrier_frequency,
@@ -483,7 +507,7 @@ class AmplitudeModulated(VoltageSource):
 
 ####################################################################################################
 
-class RandomVoltage(VoltageSource):
+class RandomMixin(SourceMixinAbc):
 
     r"""This class implements a Random Voltage source.
 
@@ -519,13 +543,8 @@ class RandomVoltage(VoltageSource):
 
     ##############################################
 
-    def __init__(self, name, node_plus, node_minus,
-                 random_type, duration=0, time_delay=0, parameter1=1, parameter2=0):
-
+    def __init__(self, random_type, duration=0, time_delay=0, parameter1=1, parameter2=0):
         # Fixme: random_type and parameters
-
-        super().__init__(name, node_plus, node_minus)
-
         self.random_type = random_type
         self.duration = as_s(duration)
         self.time_delay = as_s(time_delay)
@@ -552,3 +571,295 @@ class RandomVoltage(VoltageSource):
                 join_list((random_type, self.duration, self.time_delay,
                            self.parameter1, self.parameter2)) +
                 ')')
+
+####################################################################################################
+
+class SinusoidalVoltageSource(VoltageSource, VoltageSourceMixinAbc, SinusoidalMixin):
+
+    r"""This class implements a sinusoidal waveform voltage source.
+
+    See :class:`SinusoidalMixin` for documentation.
+
+    """
+
+    ##############################################
+
+    def __init__(self, netlist, name, node_plus, node_minus, *args, **kwargs):
+        VoltageSource.__init__(self, netlist, name, node_plus, node_minus)
+        SinusoidalMixin.__init__(self, *args, **kwargs)
+
+    ##############################################
+
+    format_spice_parameters = SinusoidalMixin.format_spice_parameters
+
+####################################################################################################
+
+class SinusoidalCurrentSource(CurrentSource, CurrentSourceMixinAbc, SinusoidalMixin):
+
+    r"""This class implements a sinusoidal waveform current source.
+
+    See :class:`SinusoidalMixin` for documentation.
+
+    """
+
+    ##############################################
+
+    def __init__(self, netlist, name, node_plus, node_minus, *args, **kwargs):
+        CurrentSource.__init__(self, netlist, name, node_plus, node_minus)
+        SinusoidalMixin.__init__(self, *args, **kwargs)
+
+    ##############################################
+
+    format_spice_parameters = SinusoidalMixin.format_spice_parameters
+
+####################################################################################################
+
+class AcLine(SinusoidalVoltageSource):
+
+    ##############################################
+
+    def __init__(self, netlist, name, node_plus, node_minus, rms_voltage=230, frequency=50):
+        super().__init__(netlist, name, node_plus, node_minus,
+                         amplitude=rms_to_amplitude(rms_voltage),
+                         frequency=frequency)
+
+####################################################################################################
+
+class PulseVoltageSource(VoltageSource, VoltageSourceMixinAbc, PulseMixin):
+
+    r"""This class implements a pulse waveform voltage source.
+
+    See :class:`PulseMixin` for documentation.
+
+    """
+
+    ##############################################
+
+    def __init__(self, netlist, name, node_plus, node_minus, *args, **kwargs):
+        VoltageSource.__init__(self, netlist, name, node_plus, node_minus)
+        PulseMixin.__init__(self, *args, **kwargs)
+
+    ##############################################
+
+    format_spice_parameters = PulseMixin.format_spice_parameters
+
+####################################################################################################
+
+class PulseCurrentSource(CurrentSource, CurrentSourceMixinAbc, PulseMixin):
+
+    r"""This class implements a pulse waveform current source.
+
+    See :class:`PulseMixin` for documentation.
+
+    """
+
+    ##############################################
+
+    def __init__(self, netlist, name, node_plus, node_minus, *args, **kwargs):
+        CurrentSource.__init__(self, netlist, name, node_plus, node_minus)
+        PulseMixin.__init__(self, *args, **kwargs)
+
+    ##############################################
+
+    format_spice_parameters = PulseMixin.format_spice_parameters
+
+####################################################################################################
+
+class ExponentialVoltageSource(VoltageSource, VoltageSourceMixinAbc, ExponentialMixin):
+
+    r"""This class implements a exponential waveform voltage source.
+
+    See :class:`ExponentialMixin` for documentation.
+
+    """
+
+    ##############################################
+
+    def __init__(self, netlist, name, node_plus, node_minus, *args, **kwargs):
+        VoltageSource.__init__(self, netlist, name, node_plus, node_minus)
+        ExponentialMixin.__init__(self, *args, **kwargs)
+
+    ##############################################
+
+    format_spice_parameters = ExponentialMixin.format_spice_parameters
+
+####################################################################################################
+
+class ExponentialCurrentSource(CurrentSource, CurrentSourceMixinAbc, ExponentialMixin):
+
+    r"""This class implements a exponential waveform current source.
+
+    See :class:`ExponentialMixin` for documentation.
+
+    """
+
+    ##############################################
+
+    def __init__(self, netlist, name, node_plus, node_minus, *args, **kwargs):
+        CurrentSource.__init__(self, netlist, name, node_plus, node_minus)
+        ExponentialMixin.__init__(self, *args, **kwargs)
+
+    ##############################################
+
+    format_spice_parameters = ExponentialMixin.format_spice_parameters
+
+####################################################################################################
+
+class PieceWiseLinearVoltageSource(VoltageSource, VoltageSourceMixinAbc, PieceWiseLinearMixin):
+
+    r"""This class implements a piece wise linear waveform voltage source.
+
+    See :class:`PieceWiseLinearMixin` for documentation.
+
+    """
+
+    ##############################################
+
+    def __init__(self, netlist, name, node_plus, node_minus, *args, **kwargs):
+        VoltageSource.__init__(self, netlist, name, node_plus, node_minus)
+        PieceWiseLinearMixin.__init__(self, *args, **kwargs)
+
+    ##############################################
+
+    format_spice_parameters = PieceWiseLinearMixin.format_spice_parameters
+
+####################################################################################################
+
+class PieceWiseLinearCurrentSource(CurrentSource, CurrentSourceMixinAbc, PieceWiseLinearMixin):
+
+    r"""This class implements a piece wise linear waveform current source.
+
+    See :class:`PieceWiseLinearMixin` for documentation.
+
+    """
+
+    ##############################################
+
+    def __init__(self, netlist, name, node_plus, node_minus, *args, **kwargs):
+        CurrentSource.__init__(self, netlist, name, node_plus, node_minus)
+        PieceWiseLinearMixin.__init__(self, *args, **kwargs)
+
+    ##############################################
+
+    format_spice_parameters = PieceWiseLinearMixin.format_spice_parameters
+
+####################################################################################################
+
+class SingleFrequencyFMVoltageSource(VoltageSource, VoltageSourceMixinAbc, SingleFrequencyFMMixin):
+
+    r"""This class implements a single frequency FM waveform voltage source.
+
+    See :class:`SingleFrequencyFMMixin` for documentation.
+
+    """
+
+    ##############################################
+
+    def __init__(self, netlist, name, node_plus, node_minus, *args, **kwargs):
+
+        VoltageSource.__init__(self, netlist, name, node_plus, node_minus)
+        SingleFrequencyFMMixin.__init__(self, *args, **kwargs)
+
+    ##############################################
+
+    format_spice_parameters = SingleFrequencyFMMixin.format_spice_parameters
+
+####################################################################################################
+
+class SingleFrequencyFMCurrentSource(CurrentSource, CurrentSourceMixinAbc, SingleFrequencyFMMixin):
+
+    r"""This class implements a single frequency FM waveform current source.
+
+    See :class:`SingleFrequencyFMMixin` for documentation.
+
+    """
+
+    ##############################################
+
+    def __init__(self, netlist, name, node_plus, node_minus, *args, **kwargs):
+        CurrentSource.__init__(self, netlist, name, node_plus, node_minus)
+        SingleFrequencyFMMixin.__init__(self, *args, **kwargs)
+
+    ##############################################
+
+    format_spice_parameters = SingleFrequencyFMMixin.format_spice_parameters
+
+####################################################################################################
+
+class AmplitudeModulatedVoltageSource(VoltageSource, VoltageSourceMixinAbc, AmplitudeModulatedMixin):
+
+    r"""This class implements a amplitude modulated waveform voltage source.
+
+    See :class:`AmplitudeModulatedMixin` for documentation.
+
+    """
+
+    ##############################################
+
+    def __init__(self, netlist, name, node_plus, node_minus, *args, **kwargs):
+        VoltageSource.__init__(self, netlist, name, node_plus, node_minus)
+        AmplitudeModulatedMixin.__init__(self, *args, **kwargs)
+
+    ##############################################
+
+    format_spice_parameters = AmplitudeModulatedMixin.format_spice_parameters
+
+####################################################################################################
+
+class AmplitudeModulatedCurrentSource(CurrentSource, CurrentSourceMixinAbc, AmplitudeModulatedMixin):
+
+    r"""This class implements a amplitude modulated waveform current source.
+
+    See :class:`AmplitudeModulatedMixin` for documentation.
+
+    """
+
+    ##############################################
+
+    def __init__(self, netlist, name, node_plus, node_minus, *args, **kwargs):
+        CurrentSource.__init__(self, netlist, name, node_plus, node_minus)
+        AmplitudeModulatedMixin.__init__(self, *args, **kwargs)
+
+    ##############################################
+
+    format_spice_parameters = AmplitudeModulatedMixin.format_spice_parameters
+
+####################################################################################################
+
+class RandomVoltageSource(VoltageSource, VoltageSourceMixinAbc, RandomMixin):
+
+    r"""This class implements a random waveform voltage source.
+
+    See :class:`RandomMixin` for documentation.
+
+    """
+
+    ##############################################
+
+    def __init__(self, netlist, name, node_plus, node_minus, *args, **kwargs):
+        VoltageSource.__init__(self, netlist, name, node_plus, node_minus)
+        RandomMixin.__init__(self, *args, **kwargs)
+
+    ##############################################
+
+    format_spice_parameters = RandomMixin.format_spice_parameters
+
+####################################################################################################
+
+class RandomCurrentSource(CurrentSource, CurrentSourceMixinAbc, RandomMixin):
+
+    r"""This class implements a random waveform current source.
+
+    See :class:`RandomMixin` for documentation.
+
+    """
+
+    ##############################################
+
+    def __init__(self, netlist, name, node_plus, node_minus, *args, **kwargs):
+        CurrentSource.__init__(self, netlist, name, node_plus, node_minus)
+        RandomMixin.__init__(self, *args, **kwargs)
+
+    ##############################################
+
+    format_spice_parameters = RandomMixin.format_spice_parameters
