@@ -35,9 +35,9 @@ import os
 
 ####################################################################################################
 
-from .BasicElement import SubCircuitElement, BipolarJunctionTransistor
+from .BasicElement import SubCircuitElement
 from .ElementParameter import FlagParameter
-from .Netlist import ElementParameterMetaClass, NPinElement, Circuit, SubCircuit
+from .Netlist import ElementParameterMetaClass, Circuit, SubCircuit
 
 ####################################################################################################
 
@@ -45,6 +45,11 @@ _module_logger = logging.getLogger(__name__)
 
 ####################################################################################################
 
+class ParseError(NameError):
+    pass
+
+####################################################################################################
+
 class PrefixData:
 
     """This class represents a device prefix."""
@@ -70,8 +75,8 @@ class PrefixData:
         self.has_optionals = has_optionals
 
         self.multi_devices = len(classes) > 1
-        self.npins = prefix in ('Q', 'X') # NPinElement, Q has 3 to 4 pins
-        if self.npins:
+        self.has_variable_number_of_pins = prefix in ('Q', 'X') # NPinElement, Q has 3 to 4 pins
+        if self.has_variable_number_of_pins:
             self.number_of_pins = None
         else:
             # Q and X are single
@@ -105,7 +110,7 @@ class PrefixData:
 ####################################################################################################
 
 _prefix_cache = {}
-for prefix, classes in ElementParameterMetaClass.__classes__.items():
+for prefix, classes in ElementParameterMetaClass._classes.items():
     prefix_data = PrefixData(prefix, classes)
     _prefix_cache[prefix] = prefix_data
     _prefix_cache[prefix.lower()] = prefix_data
@@ -154,12 +159,13 @@ class Statement:
     def __init__(self, line, statement=None):
 
         self._line = line
-        self._line.lower_case_statement(statement)
+
+        if statement is not None:
+            self._line.lower_case_statement(statement)
 
     ##############################################
 
     def __repr__(self):
-
         return '{} {}'.format(self.__class__.__name__, repr(self._line))
 
     ##############################################
@@ -183,14 +189,12 @@ class Statement:
     ##############################################
 
     def kwargs_to_python(self, kwargs):
-
         return ['{}={}'.format(key, self.value_to_python(value))
                 for key, value in kwargs.items()]
 
     ##############################################
 
     def join_args(self, args):
-
         return ', '.join(args)
 
 ####################################################################################################
@@ -209,7 +213,7 @@ class Title(Statement):
     def __init__(self, line):
 
         super().__init__(line, statement='title')
-        self._title = self._line.read_right_of('.title')
+        self._title = self._line.right_of('.title')
 
     ##############################################
 
@@ -223,6 +227,35 @@ class Title(Statement):
 
 ####################################################################################################
 
+class Lib(Statement):
+
+    """ This class implements a library definition. """
+
+    ##############################################
+
+    def __init__(self, line):
+
+        super().__init__(line, statement='lib')
+        self._lib = self._line.right_of('.lib')
+
+    ##############################################
+
+    def __str__(self):
+        return self._lib
+
+    ##############################################
+
+    def __repr__(self):
+        return 'Lib {}'.format(self._lib)
+
+    ##############################################
+
+    def to_python(self, netlist_name):
+
+        return '{}.lib({})'.format(netlist_name, self._lib) + os.linesep
+
+####################################################################################################
+
 class Include(Statement):
 
     """ This class implements a include definition. """
@@ -232,7 +265,7 @@ class Include(Statement):
     def __init__(self, line):
 
         super().__init__(line, statement='include')
-        self._include = self._line.read_right_of('.include')
+        self._include = self._line.right_of('.include').strip('"')
 
     ##############################################
 
@@ -266,12 +299,16 @@ class Model(Statement):
 
     def __init__(self, line):
 
-        super().__init__(line, statement='mode')
+        super().__init__(line, statement='model')
 
-        # Fixme
-        parameters, dict_parameters = self._line.split_line('.model')
-        self._name, self._model_type = parameters[:2]
-        self._parameters = dict_parameters
+        text = line.right_of('.model').strip()
+        import re
+        mtch = re.match('\s*([^ \t]+)\s*([^ \t(]+)(.*)', text)
+        self._name = mtch[1]
+        self._model_type = mtch[2]
+        params = mtch[3]
+        params = params.strip('() ')
+        self._parameters = Line.get_kwarg(params)
 
     ##############################################
 
@@ -283,13 +320,11 @@ class Model(Statement):
     ##############################################
 
     def __repr__(self):
-
         return 'Model {} {} {}'.format(self._name, self._model_type, self._parameters)
 
     ##############################################
 
     def to_python(self, netlist_name):
-
         args = self.values_to_python((self._name, self._model_type))
         kwargs = self.kwargs_to_python(self._parameters)
         return '{}.model({})'.format(netlist_name, self.join_args(args + kwargs)) + os.linesep
@@ -297,7 +332,6 @@ class Model(Statement):
     ##############################################
 
     def build(self, circuit):
-
         circuit.model(self._name, self._model_type, **self._parameters)
 
 ####################################################################################################
@@ -339,7 +373,6 @@ class SubCircuitStatement(Statement):
     ##############################################
 
     def __repr__(self):
-
         text = 'SubCircuit {} {}'.format(self._name, self._nodes) + os.linesep
         text += os.linesep.join(['  ' + repr(statement) for statement in self._statements])
         return text
@@ -347,18 +380,14 @@ class SubCircuitStatement(Statement):
     ##############################################
 
     def __iter__(self):
-
         """ Return an iterator on the statements. """
-
         return iter(self._statements)
 
     ##############################################
 
     def append(self, statement):
-
         """ Append a statement to the statement's list. """
-
-        self._statements .append(statement)
+        self._statements.append(statement)
 
     ##############################################
 
@@ -373,10 +402,9 @@ class SubCircuitStatement(Statement):
 
     ##############################################
 
-    def build(self, circuit):
-
-        subcircuit = SubCircuit(self._name, self._nodes)
-        SpiceParser.build_netlist(subcircuit, self._statements)
+    def build(self, ground=0):
+        subcircuit = SubCircuit(self._name, *self._nodes)
+        SpiceParser._build_circuit(subcircuit, self._statements, ground)
         return subcircuit
 
 ####################################################################################################
@@ -415,7 +443,7 @@ class Element(Statement):
         self._dict_parameters = {}
 
         # Read nodes
-        if not prefix_data.npins:
+        if not prefix_data.has_variable_number_of_pins:
             number_of_pins = prefix_data.number_of_pins
             if number_of_pins:
                 self._nodes, stop_location = self._line.read_words(stop_location, number_of_pins)
@@ -446,12 +474,13 @@ class Element(Statement):
             for kwarg in kwargs:
                 try:
                     key, value = kwarg.split('=')
-                    self._dict_parameters[key] = value
+                    self._dict_parameters[key.lower()] = value
                 except ValueError:
                     if kwarg in ('off',) and prefix_data.has_flag:
                         self._dict_parameters['off'] = True
                     else:
-                        self._logger.warn(line_str)
+                        # Fixme: warning -> debug due to spam ...
+                        self._logger.debug(line_str)
                         # raise NameError('Bad element line:', line_str)
 
         if prefix_data.multi_devices:
@@ -472,7 +501,7 @@ class Element(Statement):
         for i in to_delete:
             del self._parameters[i]
 
-        self._logger.debug(os.linesep + self.__repr__())
+        # self._logger.debug(os.linesep + self.__repr__())
 
     ##############################################
 
@@ -484,19 +513,25 @@ class Element(Statement):
     ##############################################
 
     def __repr__(self):
-
         return 'Element {0._prefix} {0._name} {0._nodes} {0._parameters} {0._dict_parameters}'.format(self)
 
+    ##############################################
+
+    def translate_ground_node(self, ground):
+
+        nodes = []
+        for node in self._nodes:
+            if str(node) == str(ground):
+                node = 0
+            nodes.append(node)
+
+        return nodes
 
     ##############################################
 
     def to_python(self, netlist_name, ground=0):
 
-        nodes = []
-        for node in self._nodes:
-            if str(node) == ground:
-                node = 0
-            nodes.append(node)
+        nodes = self.translate_ground_node(ground)
         args = [self._name]
         if self._prefix != 'X':
             args += nodes + self._parameters
@@ -510,20 +545,17 @@ class Element(Statement):
 
     def build(self, circuit, ground=0):
 
-        factory = getattr(circuit, self.factory.alias)
-        nodes = []
-        for node in self._nodes:
-            if str(node) == ground:
-                node = 0
-            nodes.append(node)
+        factory = getattr(circuit, self.factory.ALIAS)
+        nodes = self.translate_ground_node(ground)
         if self._prefix != 'X':
             args = nodes + self._parameters
         else: # != Spice
             args = self._parameters + nodes
         kwargs = self._dict_parameters
-        message = ' '.join([str(x) for x in (self._prefix, self._name, nodes,
-                                             self._parameters, self._dict_parameters)])
-        self._logger.debug(message)
+        if self._logger.isEnabledFor(logging.DEBUG):
+            message = ' '.join([str(x) for x in (self._prefix, self._name, nodes,
+                                                 self._parameters, self._dict_parameters)])
+            self._logger.debug(message)
         factory(self._name, *args, **kwargs)
 
 ####################################################################################################
@@ -532,20 +564,15 @@ class Line:
 
     """ This class implements a line in the netlist. """
 
+    _logger = _module_logger.getChild('Element')
+
     ##############################################
 
-    def __init__(self, text, line_range):
+    def __init__(self, line, line_range, end_of_line_comment):
 
-        text = str(text)
-        for marker in ('$', ';', '//'):
-            location = text.find(marker)
-            if location != -1:
-                break
-        if location != -1:
-            text = text[:location]
-            comment = text[location:]
-        else:
-            comment = ''
+        self._end_of_line_comment = end_of_line_comment
+
+        text, comment, self._is_comment = self._split_comment(line)
 
         self._text = text
         self._comment = comment
@@ -554,7 +581,7 @@ class Line:
     ##############################################
 
     def __repr__(self):
-        return '{0._line_range} {0._text}'.format(self)
+        return '{0._line_range}: {0._text} // {0._comment}'.format(self)
 
     ##############################################
 
@@ -563,24 +590,91 @@ class Line:
 
     ##############################################
 
-    def lower_case_statement(self, statement):
+    @property
+    def comment(self):
+        return self._comment
 
-        if statement:
-            self._text = self._text.replace(statement.upper(), statement.lower())
+    @property
+    def is_comment(self):
+        return self._is_comment
 
     ##############################################
 
-    def read_right_of(self, text):
+    def _split_comment(self, line):
 
+        line = str(line)
+
+        if line.startswith('*'):
+            is_comment = True
+            text = ''
+            comment = line[1:].strip()
+        else:
+            is_comment = False
+            # remove end of line comment
+            location = -1
+            for marker in self._end_of_line_comment:
+                _location = line.find(marker)
+                if _location != -1:
+                    if location == -1:
+                        location = _location
+                    else:
+                        location = min(_location, location)
+            if location != -1:
+                text = line[:location].strip()
+                comment = line[location:].strip()
+            else:
+                text = line
+                comment = ''
+
+        return text, comment, is_comment
+
+    ##############################################
+
+    def append(self, line):
+
+        text, comment, is_comment = self._split_comment(line)
+
+        if text:
+            if not self._text.endswith(' ') or text.startswith(' '):
+                self._text += ' '
+            self._text += text
+        if comment:
+            self._comment += ' // ' + comment
+
+        _slice = self._line_range
+        self._line_range = slice(_slice.start, _slice.stop + 1)
+
+    ##############################################
+
+    def lower_case_statement(self, statement):
+
+        """Lower case the statement"""
+
+        # statement without . prefix
+
+        if self._text:
+            lower_statement = statement.lower()
+            _slice = slice(1, len(statement) + 1)
+            _statement = self._text[_slice]
+            if _statement.lower() == lower_statement:
+                self._text = '.' + lower_statement + self._text[_slice.stop:]
+
+    ##############################################
+
+    def right_of(self, text):
         return self._text[len(text):].strip()
 
     ##############################################
 
     def read_words(self, start_location, number_of_words):
 
+        """Read a fixed number of words separated by space."""
+
+        words = []
+        stop_location = None
+
         line_str = self._text
         number_of_words_read = 0
-        words = []
         while number_of_words_read < number_of_words: # and start_location < len(line_str)
             stop_location = line_str.find(' ', start_location)
             if stop_location == -1:
@@ -592,9 +686,11 @@ class Line:
             if stop_location is None: # we should stop
                 if number_of_words_read != number_of_words:
                     template = 'Bad element line, looking for word {}/{}:' + os.linesep
-                    raise NameError(template.format(number_of_words_read, number_of_words) +
-                                    line_str + os.linesep +
-                                    ' '*start_location + '^')
+                    message = (template.format(number_of_words_read, number_of_words) +
+                               line_str + os.linesep +
+                               ' '*start_location + '^')
+                    self._logger.warning(message)
+                    raise ParseError(message)
             else:
                 if start_location < stop_location:
                     start_location = stop_location
@@ -607,8 +703,9 @@ class Line:
 
     def split_words(self, start_location, until=None):
 
-        line_str = self._text
         stop_location = None
+
+        line_str = self._text
         if until is not None:
             location = line_str.find(until, start_location)
             if location != -1:
@@ -626,26 +723,64 @@ class Line:
 
     ##############################################
 
+    @staticmethod
+    def get_kwarg(text):
+
+        dict_parameters = {}
+
+        parts = []
+        for part in text.split():
+            if '=' in part and part != '=':
+                left, right = [x for x in part.split('=')]
+                parts.append(left)
+                parts.append('=')
+                if right:
+                    parts.append(right)
+            else:
+                parts.append(part)
+
+        i = 0
+        i_stop = len(parts)
+        while i < i_stop:
+            if i + 1 < i_stop and parts[i + 1] == '=':
+                key, value = parts[i], parts[i + 2]
+                dict_parameters[key] = value
+                i += 3
+            else:
+                raise ParseError("Bad kwarg: {}".format(text))
+
+        return dict_parameters
+
+    ##############################################
+
     def split_line(self, keyword):
 
-        """ Split the line according to the following pattern::
+        """Split the line according to the following pattern::
 
             keyword parameter1 parameter2 ... key1=value1 key2=value2 ...
 
-        Return the list of parameters and the dictionnary.
+        Return the list of parameters and the dictionary.
+
         """
 
+        # Fixme: cf. get_kwarg
+
         parameters = []
         dict_parameters = {}
 
-        text = self._text[len(keyword):]
+        text = self.right_of(keyword)
+
         parts = []
         for part in text.split():
-            if '=' in part:
+            if '=' in part and part != '=':
                 left, right = [x for x in part.split('=')]
-                parts += [left, '=', right]
+                parts.append(left)
+                parts.append('=')
+                if right:
+                    parts.append(right)
             else:
                 parts.append(part)
+
         i = 0
         i_stop = len(parts)
         while i < i_stop:
@@ -673,27 +808,33 @@ class SpiceParser:
 
       :attr:`subcircuits`
 
+      :attr:`incl_libs`
+
     """
 
     _logger = _module_logger.getChild('SpiceParser')
 
     ##############################################
 
-    def __init__(self, path=None, source=None):
+    def __init__(self, path=None, source=None, end_of_line_comment=('$', '//', ';'), recurse=False, section=None):
 
         # Fixme: empty source
 
+        self._path = path  # For use by _parse() when recursing through files.
+
         if path is not None:
             with open(str(path), 'r') as f:
                 raw_lines = f.readlines()
         elif source is not None:
-            raw_lines = source.split(os.linesep) # Fixme: other os
+            raw_lines = source.split(os.linesep)
         else:
             raise ValueError
 
+        self._end_of_line_comment = end_of_line_comment
+
         lines = self._merge_lines(raw_lines)
         self._title = None
-        self._statements = self._parse(lines)
+        self._statements = self._parse(lines=lines, recurse=recurse, section=section)
         self._find_sections()
 
     ##############################################
@@ -705,36 +846,68 @@ class SpiceParser:
         A line starting with "+" continues the preceding line.
         """
 
-        # Fixme: better using lines[-1] ?
         lines = []
-        current_line = ''
-        current_line_index = None
-        for line_index, line in enumerate(raw_lines):
-            if line.startswith('+'):
-                current_line += ' ' + line[1:].strip()
+        current_line = None
+        for line_index, line_string in enumerate(raw_lines):
+            line_string = line_string.lstrip(' ')
+            if line_string.startswith('+'):
+                current_line.append(line_string[1:].strip('\r\n'))
             else:
-                if current_line:
-                    lines.append(Line(current_line, slice(current_line_index, line_index)))
-                current_line = line.strip()
-                current_line_index = line_index
-        if current_line:
-            lines.append(Line(current_line, slice(current_line_index, len(raw_lines))))
+                line_string = line_string.strip('\r\n')
+                if line_string:
+                    _slice = slice(line_index, line_index +1)
+                    line = Line(line_string, _slice, self._end_of_line_comment)
+                    lines.append(line)
+                    # handle case with comment before line continuation
+                    if not line_string.startswith('*'):
+                        current_line = line
+
         return lines
 
     ##############################################
 
-    def _parse(self, lines):
+    def _parse(self, lines, recurse=False, section=None):
 
         """ Parse the lines and return a list of statements. """
 
+        # The first line in the input file must be the title, which is the only comment line that does
+        # not need any special character in the first place.
+        #
+        # The last line must be .end
+
+        if len(lines) <= 1:
+            self._logger.warning('Empty Spice file: {self._path}'.format(**locals()))
+            # raise NameError('Netlist is empty')
+        # if lines[-1] != '.end':
+        #     raise NameError('".end" is expected at the end of the netlist')
+
+        title_statement = '.title '
+        self._title = str(lines[0])
+        if self._title.startswith(title_statement):
+            self._title = self._title[len(title_statement):]
+
+        # SUBCKT and MODEL files often start with their commands as the
+        # first line so they'll parse incorrectly if that line is removed.
+        # For everything else, assume the first line is a TITLE line and
+        # remove it.
+        if str(lines[0]).startswith(('.model', '.subckt')):
+            start_index = 0
+        else:
+            start_index = 1
+
         statements = []
+        skip_lines = [False]  # True on top of stack means skip lines.
         sub_circuit = None
         scope = statements
-        for line in lines:
-            # print(repr(line))
+        self.incl_libs = []  # Libraries found during recursive descent into includes.
+        for line in lines[start_index:]:
+            # print('>', repr(line))
             text = str(line)
             lower_case_text = text.lower() # !
-            if text.startswith('*'):
+            if skip_lines[-1]:
+                if lower_case_text.startswith('.endl'):
+                    skip_lines.pop()
+            elif line.is_comment:
                 scope.append(Comment(line))
             elif lower_case_text.startswith('.'):
                 lower_case_text = lower_case_text[1:]
@@ -746,6 +919,7 @@ class SpiceParser:
                     sub_circuit = None
                     scope = statements
                 elif lower_case_text.startswith('title'):
+                    # override first line
                     self._title = Title(line)
                     scope.append(self._title)
                 elif lower_case_text.startswith('end'):
@@ -754,7 +928,21 @@ class SpiceParser:
                     model = Model(line)
                     scope.append(model)
                 elif lower_case_text.startswith('include'):
-                    scope.append(Include(line))
+                    incl = Include(line)
+                    scope.append(incl)
+                    if recurse:
+                        from .Library import SpiceLibrary
+                        incl_path = os.path.join(str(self._path.directory_part()), str(incl))
+                        self.incl_libs.append(SpiceLibrary(root_path=incl_path, recurse=recurse))
+                elif lower_case_text.startswith('lib'):
+                    lib = Lib(line)
+                    if section and str(lib) != section.lower():
+                        # If the .lib statement is only followed by the name of a section,
+                        # then skip any lines in a library section whose name does not match
+                        # the library section argument.
+                        skip_lines.append(True)
+                    else:
+                        scope.append(lib)
                 else:
                     # options param ...
                     # .global
@@ -762,10 +950,14 @@ class SpiceParser:
                     # .param
                     # .func .csparam .temp .if
                     # { expr } are allowed in .model lines and in device lines.
-                    self._logger.warn(line)
+                    # self._logger.warning('Parser ignored: {}'.format(line))
+                    pass
             else:
-                element = Element(line)
-                scope.append(element)
+                try:
+                    element = Element(line)
+                    scope.append(element)
+                except ParseError:
+                    self._logger.warning('Parse error on:\n{}'.format(line))
 
         return statements
 
@@ -792,39 +984,43 @@ class SpiceParser:
     ##############################################
 
     def is_only_subcircuit(self):
-
         return bool(not self.circuit and self.subcircuits)
 
     ##############################################
 
     def is_only_model(self):
-
         return bool(not self.circuit and not self.subcircuits and self.models)
 
     ##############################################
 
     @staticmethod
-    def build_netlist(netlist, statements):
+    def _build_circuit(circuit, statements, ground):
 
         for statement in statements:
             if isinstance(statement, Include):
-                netlist.include(str(statement))
+                circuit.include(str(statement))
 
         for statement in statements:
             if isinstance(statement, Element):
-                statement.build(netlist, netlist.gnd)
+                statement.build(circuit, ground)
             elif isinstance(statement, Model):
-                statement.build(netlist)
+                statement.build(circuit)
             elif isinstance(statement, SubCircuit):
-                subcircuit = statement.build()
-                netlist.subcircuit(subcircuit)
+                subcircuit = statement.build(ground) # Fixme: ok ???
+                circuit.subcircuit(subcircuit)
 
     ##############################################
 
     def build_circuit(self, ground=0):
 
+        """Build a :class:`Circuit` instance.
+
+        Use the *ground* parameter to specify the node which must be translated to 0 (SPICE ground node).
+
+        """
+
         circuit = Circuit(str(self._title))
-        self.build_netlist(circuit, self._statements)
+        self._build_circuit(circuit, self._statements, ground)
         return circuit
 
     ##############################################
