@@ -1,24 +1,24 @@
 ####################################################################################################
 
 from PySpice.Spice.Netlist import SubCircuitFactory
-from PySpice.Unit.Units import *
+from PySpice.Unit import *
 
 ####################################################################################################
 
 class Transformer(SubCircuitFactory):
 
-    __name__ = 'Transformer'
-    __nodes__ = ('input_plus', 'input_minus',
-                 'output_plus', 'output_minus')
+    NAME = 'Transformer'
+    NODES = ('input_plus', 'input_minus',
+             'output_plus', 'output_minus')
 
     ##############################################
 
     def __init__(self,
                  turn_ratio,
-                 primary_inductance=1,
-                 copper_resistance=1,
-                 leakage_inductance=milli(1),
-                 winding_capacitance=pico(20),
+                 primary_inductance=1@u_H,
+                 copper_resistance=1@u_Ω,
+                 leakage_inductance=1@u_mH,
+                 winding_capacitance=20@u_pF,
                  coupling=.999,
              ):
 
@@ -52,9 +52,3 @@ class Transformer(SubCircuitFactory):
 
         # Coupling
         self.CoupledInductor('coupling', primary_inductor.name, secondary_inductor.name, coupling)
-
-####################################################################################################
-# 
-# End
-# 
-####################################################################################################
