@@ -1,12 +1,12 @@
 ####################################################################################################
 
-#!#
-#!# =============
-#!#  Transformer
-#!# =============
-#!#
-#!# This examples shows how to simulate a transformer.
-#!#
+#r#
+#r# =============
+#r#  Transformer
+#r# =============
+#r#
+#r# This examples shows how to simulate a transformer.
+#r#
 
 ####################################################################################################
 
@@ -27,6 +27,8 @@ from PySpice.Unit import *
 
 from Transformer import Transformer
 
+#f# literal_include('Transformer.py')
+
 ####################################################################################################
 
 circuit = Circuit('Transformer')
@@ -39,14 +41,15 @@ circuit.R('load', 'output', circuit.gnd, 1@u_kΩ)
 simulator = circuit.simulator(temperature=25, nominal_temperature=25)
 analysis = simulator.transient(step_time=ac_line.period/200, end_time=ac_line.period*3)
 
-figure = plt.figure(1, (20, 10))
-plot(analysis.input)
-plot(analysis.output)
-plt.legend(('Vin [V]', 'Vout [V]'), loc=(.8,.8))
-plt.grid()
-plt.xlabel('t [s]')
-plt.ylabel('[V]')
+figure, ax = plt.subplots(figsize=(20, 10))
+ax.plot(analysis.input)
+ax.plot(analysis.output)
+ax.legend(('Vin [V]', 'Vout [V]'), loc=(.8,.8))
+ax.grid()
+ax.set_xlabel('t [s]')
+ax.set_ylabel('[V]')
 
 plt.tight_layout()
 plt.show()
-#fig# save_figure(figure, 'transformer.png')
+
+#f# save_figure('figure', 'transformer.png')
